
package pb138xsdparser;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.*;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * The class contains methods to analyse XML Schema
 * and to bind its content to Java programming language.
 * 
 * @version 2012-06-16
 * @author Matus Durcik
 */
public class Binder {

    private Document doc;
    private XPath xpath;
    private String objectFactory;
    private String prefix;
    private ArrayList<String> classNameList = new ArrayList<String>();
    private ArrayList<String> subElementNameList;
    
    /* ---------------------------------------------------------------------- */
    
    /**
     * Create a new instance of this class
     * and read the content of the given URI.
     */
    public static Binder newInstance(URI uri)
            throws SAXException, ParserConfigurationException, IOException {
        return new Binder(uri);
    }  
    
    /**
     * Create a new instance of this class
     * and read the content of the given XML file.
     */
    public static Binder newInstance(File file)
            throws SAXException, ParserConfigurationException, IOException {
        return newInstance(file.toURI());
    }
    
    /**
     * Constructor creating a new instance of this class,
     * reading from the file at the given URI.
     */
    private Binder(URI uri)
            throws SAXException, ParserConfigurationException, IOException {
               
        DocumentBuilderFactory domFactory = DocumentBuilderFactory.newInstance();
        domFactory.setNamespaceAware(true);
        DocumentBuilder builder = domFactory.newDocumentBuilder();
        doc = builder.parse(uri.toString());
        
        XPathFactory factory = XPathFactory.newInstance();
        xpath = factory.newXPath();
        /* set namespace context */
        xpath.setNamespaceContext(new MyNamespaceContext());
   
    }
    
    /* ---------------------------------------------------------------------- */
    
    /**
     * Analyses sequences of elements within complex types
     * and creates Java variables and methods.
     * 
     * It might create an Element class instead of variable.
     * 
     * @param path XPath to elements being analysed
     * @param classWriter currently active Buffered Writer
     * @param att TRUE, if binding <xs:attribute> (not <xs:element>)
     * @param ancestorClasses string of dot separated ancestor classes (in case, it'll create an element class)
     */
    public void bindAttributes(String path, BufferedWriter classWriter, boolean att, String ancestorClasses){          
        NodeList nodeList = xPathToNodeList(path);

        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i) instanceof Element) {             
                Element element = (Element) nodeList.item(i);

                /* get name attribute */
                String nameAttribute = element.getAttribute("name");
                
                /* set name of variable */
                String variableName = nameAttribute;
                
                /* resolve name conflict, if there exists <xs:element> with same name as current <xs:attribute> */
                while(att && subElementNameList.contains(variableName)){
                    variableName += "Att";
                }
                
                /* get type of variable */
                String typeAttribute = "";
                if (element.hasAttribute("type")){
                    typeAttribute = convertType(element.getAttribute("type"));
                }else{
                    /* if current element doesn't have a type attribute, search for a base type in a restriction/extension subelement */
                    NodeList restrictionNodeList = xPathToNodeList(path + "[@name='" + nameAttribute + "']/" + prefix + ":simpleType/" + prefix + ":restriction[@base]"
                            + " | " + path + "[@name='" + nameAttribute + "']/" + prefix + ":simpleType/" + prefix + ":extension[@base]");
                    if (restrictionNodeList.item(0) instanceof Element) {
                        Element restrictionElement = (Element) restrictionNodeList.item(0);
                        typeAttribute = convertType(restrictionElement.getAttribute("base"));
                    }
                }

                /*
                 * IMPORTANT:
                 * If type attribute equals any name attribute of any <" + prefix + ":complexType> element,
                 * this element wouldn't be an attribute, but an ELEMENT CLASS!!!
                 */
                if(classNameList.contains(typeAttribute)){
                    bindElementModel(nameAttribute, typeAttribute, classWriter, ancestorClasses);
                }else{
                    /* if binding <xs:element> (not <xs:attribute>), add its name to subElementNameList*/
                    if(!att){
                        subElementNameList.add(variableName);
                    }

                    /* fixed or default? */
                    String fixedAttribute = "";
                    String defaultAttribute = "";

                    boolean isFixed = element.hasAttribute("fixed");
                    boolean isDefault = element.hasAttribute("default");

                    if(typeAttribute.equals("java.lang.String")){
                        /* we need to add "quotes", if it's String */
                        if(isFixed){
                            fixedAttribute = "\"" + element.getAttribute("fixed") + "\"";
                        }
                        if(isDefault){
                            defaultAttribute = "\"" + element.getAttribute("default") + "\"";
                        }
                    }else{
                        fixedAttribute = element.getAttribute("fixed");
                        defaultAttribute = element.getAttribute("default");
                    }


                    try{
                        /* variable definition */ 
                        classWriter.write("\tprivate " + typeAttribute + " ");                   
                        classWriter.write(variableName);
                        
                        if(isFixed){
                            classWriter.write(" = " + fixedAttribute);
                        }else{
                            if(isDefault){
                                classWriter.write(" = " + defaultAttribute);
                            }
                        }
                        classWriter.write(";\n\n");

                        /* get() method */
                        classWriter.write("\tpublic " + typeAttribute + " get" + variableName
                                + "(){\n"
                                + "\t\t" +  "return " + variableName + ";\n"
                                + "\t}\n\n");

                        /* set() method */
                        classWriter.write("\tpublic void set" + variableName
                                + "(" + typeAttribute + " input" + variableName + "){\n"
                                + "\t\t" + variableName + " = input" + variableName + ";\n"
                                + "\t}\n\n");

                    }catch(IOException e){
                        System.err.println("IOException: " + e.getMessage());
                    }
                }
            }
        }     
    }
    
    /* ---------------------------------------------------------------------- */
    
    /**
     * Complex type elements are recursively analysed
     * and bound to Java content classes.
     * 
     * @param path XPath of parent element
     * @param ancestorClasses string of dot separated ancestor classes (e.g. "OuterClass.InnerClass")
     * @param classWriter currently active Buffered Writer
     */
    public void bindContentModels(String path, String ancestorClasses, BufferedWriter classWriter){
        String newAncestorClasses = ancestorClasses;
        String newPath;
        NodeList nodeList;
       
        /*
         *  ##### complexTypeStyle = 0 ##### 
         * 
         *  <xs:complexType name="name">
         *    ...
         *  </xs:complexType>
         * 
         * 
         *  ##### complexTypeStyle = 1 #####   
         * 
         *  <xs":element name="name">
         *    <xs:complexType>
         *    ...
         *    </xs:complexType>
         *  </xs":element>
         *
         */
        for (int complexTypeStyle = 0; complexTypeStyle < 2; complexTypeStyle++){
            
            if (complexTypeStyle == 0){
                newPath = path + "/" + prefix + ":complexType[@name]";    
            }else{
                newPath = path + "/" + prefix + ":element[@name and " + prefix + ":complexType]";   
            }
            
            nodeList = xPathToNodeList(newPath); 
            
            if (nodeList != null){
                for (int i = 0; i < nodeList.getLength(); i++) {
                    if (nodeList.item(i) instanceof Element) {             
                        Element element = (Element) nodeList.item(i);

                        /* get name attribute */
                        String nameAttribute = element.getAttribute("name");
                        
                        /* set name of class */
                        String className = nameAttribute;
                        int j = 1;
                       
                        while (classNameList.contains(className)){
                            /* resolve name conflicts */
                            className = nameAttribute + j;
                            j++;
                        }
                        if (complexTypeStyle == 0 && j > 1){
                            /* ak je complexTypeStyle 0, musime zaobstarat zmenu aj typov vsetkych sibling element, ktore maju meno tohto complexType vo svojom type */
                            String siblingPath = path + "/" + prefix + ":element[@type='" + nameAttribute + "']";
                            NodeList siblingNodeList = xPathToNodeList(siblingPath);
                            if (siblingNodeList != null){
                                for (int k = 0; k < siblingNodeList.getLength(); k++) {
                                    Element siblingElement = (Element) siblingNodeList.item(k);
                                    siblingElement.setAttribute("type", nameAttribute + (j - 1));
                                }
                            }
                        }
                        
                        try{
                            /* add name to classNameList */
                            classNameList.add(className);
                            
                            /* factory method */
                            createFactoryMethod(className, ancestorClasses);

                            /* new file & buffered writer & package definition (if it's root) */
                            if (ancestorClasses.isEmpty()) {
                                File classFile = new File("generated/" + className + ".java");
                                classWriter = new BufferedWriter(new FileWriter(classFile));
                                /* package definition */
                                classWriter.write("package generated;\n\n");
                            }

                            /* class definition */
                            classWriter.write("public class " + className + " {\n\n");

                            /* add class to new ancestor classes string */
                            if (ancestorClasses.isEmpty()){
                                newAncestorClasses = className;
                            }else{
                                newAncestorClasses = ancestorClasses + "." + className;
                            }
                                                        
                            subElementNameList = new ArrayList<String>();
                            
                            if (complexTypeStyle == 0){
                                /* nested classes */
                                bindContentModels(newPath + "[@name='" + nameAttribute + "']/*", newAncestorClasses, classWriter);

                                /* attributes & get/set methods */
                                bindAttributes(newPath + "[@name='" + nameAttribute + "']/*/" + prefix + ":element[@name][@type or " + prefix + ":simpleType/*[@base]]", classWriter, false, newAncestorClasses);
                                bindAttributes(newPath + "[@name='" + nameAttribute + "']/" + prefix + ":attribute[@name][@type or " + prefix + ":simpleType/*[@base]]", classWriter, true, newAncestorClasses);
                            }else{
                                /* nested classes */
                                bindContentModels(path + "/" + prefix + ":element[@name='" + nameAttribute + "']/" + prefix + ":complexType/*", newAncestorClasses, classWriter);

                                /* attributes & get/set methods */
                                bindAttributes(path + "/" + prefix + ":element[@name='" + nameAttribute + "']/" + prefix + ":complexType/*/" + prefix + ":element[@name][@type or " + prefix + ":simpleType/*[@base]]", classWriter, false, newAncestorClasses);
                                bindAttributes(path + "/" + prefix + ":element[@name='" + nameAttribute + "']/" + prefix + ":complexType/" + prefix + ":attribute[@name][@type or " + prefix + ":simpleType/*[@base]]", classWriter, true, newAncestorClasses); 
                            }
                            
                            classWriter.write("}\n\n");
                            
                            /* close buffered writer (if it's root) */
                            if (ancestorClasses.isEmpty()) {
                                classWriter.close();
                            }

                        }catch(IOException e){
                            System.err.println("IOException: " + e.getMessage());
                        }catch(NullPointerException e){
                            System.err.println("NullPointerException: " + e.getMessage());
                        }       
                    }
                }
            }
        }       
    }
    
    /* ---------------------------------------------------------------------- */
    
    /**
     * An element class is created for given name and type.
     * 
     * @param path XPath of parent element
     */
    public void bindElementModel(String nameAttribute, String typeAttribute, BufferedWriter classWriter, String ancestorClasses){
        try{
            /* set name of class */
            String className = nameAttribute;
            int j = 1;
            while (classNameList.contains(className)){
                className = nameAttribute + j;
                j++;
            }
            /* add name to classNameList */
            classNameList.add(className);
            /* create factory method */
            createFactoryMethod(className, ancestorClasses);
            /* class declaration */
            classWriter.write("public class " + className + " extends " + typeAttribute + " {\n\n}\n\n");
        }catch(IOException e){
            System.err.println("IOException: " + e.getMessage());
        }

    }
    
    /* ---------------------------------------------------------------------- */
    
    /**
     * It starts to analyse input XML and opens new Buffered File Writer.
     * Create ObjectFactory class.
     * 
     * @param output Output file
     */
    public void analyseXML(){
        prefix = doc.getDocumentElement().getPrefix();
        try{
            new File("generated").mkdir();
            File objectFactoryFile = new File("generated/ObjectFactory.java");
            
            BufferedWriter objectFactoryWriter = new BufferedWriter(new FileWriter(objectFactoryFile));
            
            objectFactory = ("package generated;\n\n"
                    + "/**\n"
                    + " * The class contains a class factory method\n"
                    + " * for each Java content and element class within the package.\n"
                    + " *\n"
                    + " */\n"
                    + "public class ObjectFactory {\n\n");
            
            bindContentModels("/" + prefix + ":schema", "", null);
            
            objectFactory += "\n}\n"; // end of ObjectFactory
            objectFactoryWriter.write(objectFactory);
            objectFactoryWriter.close();
        }catch(IOException e){
                System.err.println("IOException: " + e.getMessage());
        }
    }

    /* ---------------------------------------------------------------------- */
    
    /**
     * It converts XML attribute simple type to Java type.
     * 
     * @param typeXML XML attribute type
     * @return corresponding Java type - if corresponding type exists;
     *         original type otherwise
     */
    public String convertType(String typeXML){
        
        String typeJava;
        String DEFAULT_TYPE_JAVA = "java.lang.String";
        
        HashMap<String, String> typeMap = new HashMap<String, String>(); 
        
        typeMap.put(prefix + ":string", "java.lang.String");
        typeMap.put(prefix + ":integer", "java.math.BigInteger");
        typeMap.put(prefix + ":int", "int");
        typeMap.put(prefix + ":long", "long");
        typeMap.put(prefix + ":short", "short");
        typeMap.put(prefix + ":decimal", "java.math.BigDecimal");
        typeMap.put(prefix + ":float", "float");
        typeMap.put(prefix + ":double", "double");
        typeMap.put(prefix + ":boolean", "boolean");
        typeMap.put(prefix + ":byte", "byte");
        typeMap.put(prefix + ":QName", "javax.xml.namespace.QName");
        typeMap.put(prefix + ":dateTime", "java.lang.String");    //typeMap.put(prefix + ":dateTime", "java.util.Calendar");
        typeMap.put(prefix + ":base64Binary", "byte[]");
        typeMap.put(prefix + ":hexBinary", "byte[]");
        typeMap.put(prefix + ":unsignedInt", "long");
        typeMap.put(prefix + ":unsignedShort", "int");
        typeMap.put(prefix + ":unsignedByte", "short");
        typeMap.put(prefix + ":time", "java.lang.String");    //typeMap.put(prefix + ":time", "java.util.Calendar");
        typeMap.put(prefix + ":date", "java.lang.String");    //typeMap.put(prefix + ":date", "java.util.Calendar");
        typeMap.put(prefix + ":anyURI", "java.lang.String");
        typeMap.put(prefix + ":anySimpleType", "java.lang.String");
        
        if(typeMap.containsKey(typeXML)){
            typeJava = typeMap.get(typeXML);
        }else{
            typeJava = typeXML;
        }
        
        return typeJava;
    }

    /* ---------------------------------------------------------------------- */    
    
    /**
     * It creates Factory method for given class name attribute.
     * 
     * e.g. (if there are 2 ancestor (outer) classes)
     * 
     * Outer1.Outer2.Inner createInner(){   
     *      Outer1.Outer2 Outer2Object = createOuter2();
     *      return Outer2Object.new Inner();
     * }
     * 
     * @param nameAttribute class name attribute
     * @param ancestorClasses string of dot separated ancestor classes (e.g. "OuterClass.InnerClass")
     */
    public void createFactoryMethod(String className, String ancestorClasses){
        if (ancestorClasses.isEmpty()){
            objectFactory += "\tpublic " + className
                    + " " + "create" + className + "(){\n"
                    + "\t\treturn new " + className + "();\n"
                    + "\t}\n\n";
        }else{ 
            try{
                String[] ancestorClassesArray = ancestorClasses.split("\\.");
                String parentClass = ancestorClassesArray[ancestorClassesArray.length - 1];

                objectFactory += "\tpublic " + ancestorClasses + "." + className
                        + " " + "create" + className + "(){\n";

                objectFactory += "\t\t" + ancestorClasses + " " + parentClass + "Object"
                            + " = create" + parentClass + "();\n";

                objectFactory += "\t\treturn "
                        + parentClass + "Object.new "
                        + className + "();\n"
                        + "\t}\n\n";
            }catch(ArrayIndexOutOfBoundsException e){
                 System.err.println("ArrayIndexOutOfBoundsException: " + e.getMessage());
            }
        }
    }
    
    /* ---------------------------------------------------------------------- */   
    
    /**
     * It returns a node list for given XPath.
     * 
     * @param path XPath
     * @return Node list
     */
    public NodeList xPathToNodeList(String path){
        NodeList nodeList = null;
        try{
            XPathExpression expr = xpath.compile(path);
            Object result = expr.evaluate(doc, XPathConstants.NODESET); 
            nodeList = (NodeList) result;          
        }catch(XPathExpressionException e){
            System.err.println("XPathExpressionException: " + e.getMessage());
        }
        return nodeList;
    }
    
} /* end of Binder class */