package pb138xsdparser;

import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import javax.xml.parsers.*;
import javax.xml.xpath.*;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * This class creates a parser, which transforms XML document to Java Objects.
 *
 * @author Petr Machovec
 * @version 2012-06-15
 */
public class ParserMaker {
    private File xmlSchema;
    private Document doc;
    private String parserName;
    private String prefix = "";
    private List<String> classNameList = new ArrayList<String>();
   
    public ParserMaker(File xmlSchema){
        this.xmlSchema = xmlSchema;
    }    
   
    /**
     * Creates the parser for creating class instances of the XML file
     *
     * @throws ParserConfigurationException
     * @throws SAXException
     * @throws IOException
     * @throws XPathExpressionException
     */
    public void makeParser() throws ParserConfigurationException, SAXException, IOException, XPathExpressionException{
        DocumentBuilderFactory documentFactory = DocumentBuilderFactory.newInstance();
        documentFactory.setNamespaceAware(true);
        DocumentBuilder documentBuilder = documentFactory.newDocumentBuilder();
        doc = documentBuilder.parse(xmlSchema);
       
        XPathFactory xPathFactory = XPathFactory.newInstance();
        XPath xpath = xPathFactory.newXPath();
        XPathExpression expr = xpath.compile("/*/*");
       
        Object result = expr.evaluate(doc, XPathConstants.NODESET);
       
        NodeList nodes = (NodeList) result;
        Boolean strangeSchema = false;
        if (nodes.getLength() < 1){
            throw new IOException("Bad XML schema");
        }
        if (nodes.getLength() > 1){
            strangeSchema = true;
        }
       
        Node node = nodes.item(0);
        NamedNodeMap attributes = node.getAttributes();
        String name = attributes.getNamedItem("name").getNodeValue();
        parserName = name + "parser";
       
        new File("generated").mkdir();
       
        if (!(new File("generated").exists())){
            throw new IOException("Folder for parser wasn't created");
        }
       
        File parserFile = new File("generated/" + parserName + ".java");
        parserFile.createNewFile();
       
        if (!(new File("generated/" + parserName + ".java").exists())){
            throw new IOException("Parser file wasn't created");
        }
       
        prefix = node.getPrefix();
       
        BufferedWriter out = new BufferedWriter(new FileWriter("generated/" + parserName + ".java"));
        String fileContent = "package generated;\n" +
                             "\n" +
                             "import generated.*;\n" +
                             "import java.io.*;\n" +
                             "import java.util.*;\n"+
                             "import javax.xml.parsers.*;\n" +
                             "import javax.xml.xpath.*;\n" +
                             "import org.w3c.dom.*;\n" +
                             "import org.xml.sax.*;\n" +
                             "\n" +
                             "public class " + parserName + " {\n" +
                             "    private generated.ObjectFactory objectFactory = new generated.ObjectFactory();\n" +
                             "    \n" +
                             "    /**\n" +
                             "    * Converts given XML to instances of classes made before.\n" +
                             "    *\n" +
                             "    * @param file Input XML file\n" +
                             "    * @throws ParserConfigurationException\n" +
                             "    * @throws SAXException\n" +
                             "    * @throws IOException\n" +
                             "    * @throws XPathExpressionException\n" +
                             "    */\n" +
                             "    public void parse(File file) throws ParserConfigurationException, SAXException, IOException, XPathExpressionException {\n" +
                             "        DocumentBuilderFactory documentFactory = DocumentBuilderFactory.newInstance();\n" +
                             "        documentFactory.setNamespaceAware(true);\n" +
                             "        DocumentBuilder documentBuilder = documentFactory.newDocumentBuilder();\n" +
                             "        Document doc = documentBuilder.parse(file);\n" +
                             "        \n" +
                             "        XPathFactory xPathFactory = XPathFactory.newInstance();\n" +
                             "        XPath xpath = xPathFactory.newXPath();\n" +
                             "        XPathExpression expr = xpath.compile(\"/*\");\n" +
                             "        Object result = expr.evaluate(doc, XPathConstants.NODESET);\n" +
                             "        \n" +
                             "        NodeList nodes = (NodeList) result;\n" +
                             "        if (nodes.getLength() > 1){\n" +
                             "            throw new SAXException(\"XML has more than one root element\");\n" +
                             "        }\n" +
                             "        if (!(nodes.item(0).getNodeName().equals(\"" + name + "\"))){\n" +
                             "            throw new SAXException(\"XML doesn't match given schema\");\n" +
                             "        }\n" +
                             "        \n" +
                             "        Node root = nodes.item(0);\n" +
                             "        " + name + "Instance = instance" + name + "(root);\n" +
                             "    }\n" +
                             "    \n";
       
        String way = "generated";
        if (isComplex(node)){
            fileContent = createInstanceMethod(node, fileContent, way);
        } else{
            if (isStrange(node)){
                Node complexNode = getStrangeComplex(node);
                List<Node> strangeBrothers = getStrangeBrothers(complexNode);
                fileContent = createStrangeInstanceMethod(node, fileContent, way, complexNode);
            }
        }
       
        fileContent = fileContent +
                             "    private NodeList underElements(Node node) throws XPathExpressionException {\n" +
                             "        XPathFactory xPathFactory = XPathFactory.newInstance();\n" +
                             "        XPath xpath = xPathFactory.newXPath();\n" +
                             "        XPathExpression expr = xpath.compile(\"*\");\n" +
                             "        Object result = expr.evaluate(node, XPathConstants.NODESET);\n" +
                             "        \n" +
                             "        NodeList nodes = (NodeList) result;\n" +
                             "        \n" +
                             "        return nodes;\n" +
                             "    }\n" +
                             "\n";
       
        fileContent = fileContent +
                             "\n" +
                             "}";
       
        out.write(fileContent);
        out.close();
    }
   
   
    /**
     * Creates method for making instances of given complex element
     *
     * @param node
     * @param fileContent
     * @return Text content of the output file with created method
     * @throws IOException
     * @throws XPathExpressionException
     */
    private String createInstanceMethod(Node node, String fileContent, String way) throws IOException, XPathExpressionException{
        NamedNodeMap attributes = node.getAttributes();
        String name = attributes.getNamedItem("name").getNodeValue();
        if (name == null){
            throw new IOException("Bad XML schema");
        }
        classNameList.add(name);
        if (getOccurence(name, classNameList) > 1){
            name = name + (getOccurence(name, classNameList) - 1);
        }
        
        
        List<String> attributeNamesList = getAttributes(node);
        List<Node> underElementsList = getUnderElements(node);
       
        fileContent = fileContent +
                             "    /**\n" +
                             "    /* Creates an instance of a complex element in given XML\n" +
                             "    /*\n" +
                             "    /* @param node Element in XML\n" +
                             "    */\n" +
                             "    public " + way + "." + name + " instance" + name + "(Node node) throws XPathExpressionException {\n";                            
       
        if (way.equals("generated")){
            fileContent = fileContent +
                             "        " + name + "Instance = objectFactory.create" + name + "();\n";
        } else{
            fileContent = fileContent +
                             "        " + way + "." + name + " " + name + "Instance = objectFactory.create" + name + "();\n";
        }
       
        fileContent = fileContent +
                             "        \n" +
                             "        NodeList " + name + "UnderElements = underElements(node);\n" +
                             "        for (int i = 0; i < " + name + "UnderElements.getLength(); i++) {\n";
       
        for (Node element: underElementsList){
            String elementName = element.getAttributes().getNamedItem("name").getNodeValue();
           
            fileContent = fileContent +
                             "            if (" + name + "UnderElements.item(i).getNodeName().equals(\"" + elementName + "\")){\n";
           
            if (isComplex(element) || isStrange(element)){
                if (classNameList.contains(elementName)){
                    elementName = elementName + (getOccurence(elementName, classNameList));
                }
               
                fileContent = fileContent +
                             "                " + elementName + "List.add(instance" + elementName +"(" + name + "UnderElements.item(i)));\n";                            
            } else{
                if (isSimple(element)){
                    Boolean typeSet = false;
                    String elementType = getType(element);

                    if (elementType.equals("java.lang.String")){
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(" + name + "UnderElements.item(i).getTextContent());\n";
                        typeSet = true;
                    }

                    if (elementType.equals("java.math.BigInteger")){
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(new java.math.BigInteger(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("int")){
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(Integer.parseInt(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("long")){
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(Long.parseLong(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("short")){
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(Short.parseShort(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("java.math.BigDecimal")){
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(new java.math.BigDecimal(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("float")){
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(Float.parseFloat(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("double")){
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(Double.parseDouble(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("boolean")){
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(Boolean.getBoolean(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("byte")){
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(Byte.parseByte(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("javax.xml.namespace.QName")){
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(javax.xml.namespace.QName.valueOf(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("byte[]")){
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(" + name + "UnderElements.item(i).getTextContent().getBytes());\n";
                        typeSet = true;
                    }

                    if (!(typeSet)){                    
                        fileContent = fileContent +
                                 "                " + name + "Instance.set" + elementName + "(" + name + "UnderElements.item(i).getTextContent());\n";                    
                    }
                }
            }
           
            fileContent = fileContent +
                             "            }\n" +
                             "            \n";
        }
       
        fileContent = fileContent +
                             "        }\n" +
                             "        \n";
       
        if (!(attributeNamesList.isEmpty())){
            List<String> underElementsNames = new ArrayList<String>();
            for (Node underElement: underElementsList){
                underElementsNames.add(underElement.getAttributes().getNamedItem("name").getNodeValue());
            }
            for(String attribute: attributeNamesList){
                while (underElementsNames.contains(attribute)){
                    attribute = attribute + "Att";
                }
                fileContent = fileContent +
                             "        " + name + "Instance.set" + attribute + "(node.getAttributes().getNamedItem(\"" + attribute + "\").getNodeValue());\n";
            }
        }
       
        fileContent = fileContent +                            
                             "        return " + name + "Instance;\n" +
                             "    }\n" +
                             "\n";
        if (way.equals("generated")){
            fileContent = fileContent +
                             "    private " + way + "." + name + " " + name + "Instance = new " + way + "." + name + "(); // Instance of the root element\n" +
                             "\n" +
                             "    public " + way + "." + name + " get" + name + "Instance() {\n" +
                             "        return " + name + "Instance;\n" +
                             "    }\n" +
                             "\n";
        } else{
            fileContent = fileContent +
                             "    private List<" + way + "." + name + "> " + name + "List = new ArrayList<" + way + "." + name + ">(); // List of instances of class " + name + "\n" +
                             "\n" +
                            "    public List<" + way + "." + name + "> " + " get" + name + "List() {\n" +
                             "        return " + name + "List;\n" +
                             "    }\n" +
                             "\n";
        }
                             
       
        way = way + "." + name;
       
        for (Node underElement: underElementsList){
            if (isComplex(underElement)){
                fileContent = createInstanceMethod(underElement,fileContent,way);
            } else{
                if (isStrange(underElement)){
                    Node complexUnderNode = getStrangeComplex(underElement);
                    List<Node> underStrangeBrothers = getStrangeBrothers(complexUnderNode);
                    fileContent = createStrangeInstanceMethod(underElement, fileContent, way, complexUnderNode);
                }
            }
        }
       
        return fileContent;
    }
    
    
    /**
     * Creates method for making instances of given element and all his brothers with reference to the complex type
     *
     * @param node
     * @param fileContent
     * @return Text content of the output file with created method
     * @throws IOException
     * @throws XPathExpressionException
     */
    private String createStrangeInstanceMethod(Node node, String fileContent, String way, Node complexNode) throws IOException, XPathExpressionException{
        List<String> attributeNamesList = getAttributes(complexNode);
        List<Node> underElementsList = getUnderElements(complexNode);
        
        String name = node.getAttributes().getNamedItem("name").getNodeValue();
        if (name == null){
            throw new IOException("Bad XML schema");
        }
        classNameList.add(name);
        if (getOccurence(name, classNameList) > 1){
            name = name + (getOccurence(name, classNameList) - 1);
        }

        fileContent = fileContent +
                             "    /**\n" +
                             "    /* Creates an instance of a complex element in given XML\n" +
                             "    /*\n" +
                             "    /* @param node Element in XML\n" +
                             "    */\n" +
                             "    public " + way + "." + name + " instance" + name + "(Node node) throws XPathExpressionException {\n";                            
       
        if (way.equals("generated")){
            fileContent = fileContent +
                             "        " + name + "Instance = objectFactory.create" + name + "();\n";
        } else{
            fileContent = fileContent +
                             "        " + way + "." + name + " " + name + "Instance = objectFactory.create" + name + "();\n";
        }

        fileContent = fileContent +
                             "        \n" +
                             "        NodeList " + name + "UnderElements = underElements(node);\n" +
                             "        for (int i = 0; i < " + name + "UnderElements.getLength(); i++) {\n";
       
        for (Node element: underElementsList){
            String elementName = element.getAttributes().getNamedItem("name").getNodeValue();

            fileContent = fileContent +
                             "            if (" + name + "UnderElements.item(i).getNodeName().equals(\"" + elementName + "\")){\n";
           
            if (isComplex(element) || isStrange(element)){
                if (classNameList.contains(elementName)){
                    elementName = elementName + (getOccurence(elementName, classNameList));
                }

                fileContent = fileContent +
                             "                " + elementName + "List.add(instance" + elementName +"(" + name + "UnderElements.item(i)));\n";                            
            } else{
                if (isSimple(element)){
                    Boolean typeSet = false;
                    String elementType = getType(element);

                    if (elementType.equals("java.lang.String")){
                            fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(" + name + "UnderElements.item(i).getTextContent());\n";
                        typeSet = true;
                    }

                    if (elementType.equals("java.math.BigInteger")){
                        fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(new java.math.BigInteger(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("int")){
                        fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(Integer.parseInt(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("long")){
                        fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(Long.parseLong(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("short")){
                            fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(Short.parseShort(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("java.math.BigDecimal")){
                            fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(new java.math.BigDecimal(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("float")){
                            fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(Float.parseFloat(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("double")){
                            fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(Double.parseDouble(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("boolean")){
                            fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(Boolean.getBoolean(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("byte")){
                            fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(Byte.parseByte(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("javax.xml.namespace.QName")){
                            fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(javax.xml.namespace.QName.valueOf(" + name + "UnderElements.item(i).getTextContent()));\n";
                        typeSet = true;
                    }

                    if (elementType.equals("byte[]")){
                            fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(" + name + "UnderElements.item(i).getTextContent().getBytes());\n";
                        typeSet = true;
                    }

                    if (!(typeSet)){                    
                        fileContent = fileContent +
                             "                " + name + "Instance.set" + elementName + "(" + name + "UnderElements.item(i).getTextContent());\n";                    
                    }
                }
            }

            fileContent = fileContent +
                             "            }\n" +
                             "            \n";
            }
            fileContent = fileContent +
                             "        }\n" +
                             "        \n";
       
            if (!(attributeNamesList.isEmpty())){
                List<String> underElementsNames = new ArrayList<String>();
                for (Node underElement: underElementsList){
                    underElementsNames.add(underElement.getAttributes().getNamedItem("name").getNodeValue());
                }
                for(String attribute: attributeNamesList){
                    while (underElementsNames.contains(attribute)){
                        attribute = attribute + "Att";
                    }
                fileContent = fileContent +
                             "        " + name + "Instance.set" + attribute + "(node.getAttributes().getNamedItem(\"" + attribute + "\").getNodeValue());\n";
                }
            }

        fileContent = fileContent +                            
                             "        return " + name + "Instance;\n" +
                             "    }\n" +
                             "\n";
        if (way.equals("generated")){
            fileContent = fileContent +
                             "    private " + way + "." + name + " " + name + "Instance = new " + way + "." + name + "(); // Instance of the root element\n" +
                             "\n" +
                             "    public " + way + "." + name + " get" + name + "Instance() {\n" +
                             "        return " + name + "Instance;\n" +
                             "    }\n" +
                             "\n";
        } else{
            fileContent = fileContent +
                             "    private List<" + way + "." + name + "> " + name + "List = new ArrayList<" + way + "." + name + ">(); // List of instances of class " + name + "\n" +
                             "\n" +
                             "    public List<" + way + "." + name + "> " + " get" + name + "List() {\n" +
                             "        return " + name + "List;\n" +
                             "    }\n" +
                             "\n";
        }


        way = way + "." + name;

        for (Node underElement: underElementsList){
            if (isComplex(underElement)){
                fileContent = createInstanceMethod(underElement,fileContent,way);
            } else{

                if (isStrange(underElement)){
                    System.out.println(underElement.getAttributes().getNamedItem("name").getNodeValue());
                    Node complexUnderNode = getStrangeComplex(node);
                    List<Node> underStrangeBrothers = getStrangeBrothers(complexUnderNode);
                    fileContent = createStrangeInstanceMethod(underElement, fileContent, way, complexUnderNode);                    
                }
            }
        }
        
        return fileContent;
    }
   
   
    /**
     * Determines whether the given element is complex
     *
     * @param node The element to be determined to be complex
     * @return Whether the given element is complex or not
     * @throws XPathExpressionException
     */
    private boolean isComplex(Node node) throws XPathExpressionException{
        XPathFactory xPathFactory = XPathFactory.newInstance();
        XPath xpath = xPathFactory.newXPath();
        XPathExpression expr = xpath.compile("*");
       
        Object result = expr.evaluate(node, XPathConstants.NODESET);
       
        NodeList nodes = (NodeList) result;
        if (nodes.getLength() == 0){
            return false;
        }
       
        if (nodes.item(0).getNodeName().equals(prefix + ":complexType")){
            if (nodes.getLength() > 1){
                return false;
            }
           
            node = nodes.item(0);
            expr = xpath.compile("*");
            result = expr.evaluate(node, XPathConstants.NODESET);
       
            nodes = (NodeList) result;
            if (!((nodes.item(0).getNodeName().equals(prefix + ":sequence")) || (nodes.item(0).getNodeName().equals(prefix + ":all")) || (nodes.item(0).getNodeName().equals(prefix + ":choice")))){
                return false;
            }
            return true;
        }
        return false;
    }
    
    /**
     * Determines whether the given element has a reference to a complex type
     *
     * @param node The element to be determined to be complex
     * @return Whether the given element is has a reference or not
     * @throws XPathExpressionException
     */
    private Boolean isStrange(Node node) throws XPathExpressionException{
        XPathFactory xPathFactory = XPathFactory.newInstance();
        XPath xpath = xPathFactory.newXPath();
        XPathExpression expr = xpath.compile("../*");
       
        Object result = expr.evaluate(node, XPathConstants.NODESET);
        
        NodeList nodes = (NodeList) result;
        List<String> nodesNames = new ArrayList();
        int complexPosition = -1;
        int complexNumber = 0;
        for (int i = 0; i < nodes.getLength(); i++){
            nodesNames.add(nodes.item(i).getNodeName());
            if (nodes.item(i).getNodeName().equals(prefix + ":complexType")){
                complexPosition = i;
                complexNumber++;
            }
        }
        if ((nodes.getLength() == 0) || (complexPosition == (-1)) || (complexNumber > 1)){
            return false;
        } else{
            Node complex = nodes.item(complexPosition);
            String complexName = "";
            
            if (complex.getAttributes().getNamedItem("name") == null){
                return false;
            } else {
                complexName = complex.getAttributes().getNamedItem("name").getNodeValue();
            }
            
            for (int i = 0; i < (nodes.getLength()); i++){
                if (!(nodes.item(i).equals(complex))){
                    if ((nodes.item(i).getAttributes().getNamedItem("type") == null) || (!((nodes.item(i).getAttributes().getNamedItem("type").getNodeValue().equals(complexName))))){
                        return false;
                    }
                }
            }
            
            return true;
        }
    }
    
    /**
     * Determines whether the given element is simple.
     * 
     * @param node
     * @return Whether the element is simple or not
     * @throws XPathExpressionException 
     */
    private Boolean isSimple(Node node) throws XPathExpressionException{
        XPathFactory xPathFactory = XPathFactory.newInstance();
        XPath xpath = xPathFactory.newXPath();
        XPathExpression expr = xpath.compile("*");
       
        Object result = expr.evaluate(node, XPathConstants.NODESET);
        
        NodeList nodes = (NodeList) result;
        
        if ((nodes.getLength() == 0) || (nodes.item(0).getNodeName().equals(prefix + ":simpleType"))){
            return true;
        } else{
            return false;
        }
    }
   
    /**
     * Finds attributes of the given element
     *
     * @param node
     * @return List of found attributes names. If the element has no attributes, it returns empty list.
     * @throws XPathExpressionException
     */
    private List<String> getAttributes(Node node) throws XPathExpressionException {
        XPathFactory xPathFactory = XPathFactory.newInstance();
        XPath xpath = xPathFactory.newXPath();
        XPathExpression expr = xpath.compile("*");
        Object result = expr.evaluate(node, XPathConstants.NODESET);
        NodeList nodes = (NodeList) result;
       
        Node node2 = nodes.item(0);
       
        if (node.getNodeName().equals(prefix + ":element")){
            expr = xpath.compile("*");
            result = expr.evaluate(node2, XPathConstants.NODESET);
            nodes = (NodeList) result;
       
            node2 = nodes.item(0);
        }
        
       
        expr = xpath.compile("following-sibling::*");
        result = expr.evaluate(node2, XPathConstants.NODESET);
        nodes = (NodeList) result;
       
        List<String> attributeList = new ArrayList<String>();
       
        if (nodes.getLength() == 0){
            return attributeList;
        } else{
            for (int i = 0; i < nodes.getLength(); i++){
                if (nodes.item(i).getNodeName().equals(prefix + ":attribute")){
                    NamedNodeMap attributeNameAndOther = nodes.item(i).getAttributes();
                    attributeList.add(attributeNameAndOther.getNamedItem("name").getNodeValue());
                }
            }
            return attributeList;
        }
    }

    /**
     * Gets all the elements that are children of the given complex node
     *
     * @param node
     * @return List of all children of the given complex node
     * @throws XPathExpressionException
     */
    private List<Node> getUnderElements(Node node) throws XPathExpressionException, IOException {
        XPathFactory xPathFactory = XPathFactory.newInstance();
        XPath xpath = xPathFactory.newXPath();
        XPathExpression expr = xpath.compile("*");
        Object result = expr.evaluate(node, XPathConstants.NODESET);
        NodeList nodes = (NodeList) result;
       
        Node node2 = nodes.item(0);
       
        if (node.getNodeName().equals(prefix + ":element")){
            expr = xpath.compile("*");
            result = expr.evaluate(node2, XPathConstants.NODESET);
            nodes = (NodeList) result;
       
            node2 = nodes.item(0);
        }
       
        expr = xpath.compile("child::*");
        result = expr.evaluate(node2, XPathConstants.NODESET);
        nodes = (NodeList) result;
       
        List<Node> underElementsNodes = new ArrayList<Node>();
       
        if (nodes.getLength() == 0){
            return underElementsNodes;
        } else{
            for (int i = 0; i < nodes.getLength(); i++){
                if (nodes.item(i).getNodeName().equals(prefix + ":element")){
                    underElementsNodes.add(nodes.item(i));
                }
            }
            return underElementsNodes;
        }
    }

    /**
     * Gets the value of attribute 'type' of given simple element, which is located directly in the element declaration
     * or in in it's ancestor.
     *
     * @param element
     * @return The value of attribute 'type'
     * @throws XPathExpressionException
     */
    private String getType(Node element) throws XPathExpressionException {
        NamedNodeMap attributes = element.getAttributes();
        String type = "";
        if (attributes.getNamedItem("type") != null){
            type = attributes.getNamedItem("type").getNodeValue();
        } else{
            XPathFactory xPathFactory = XPathFactory.newInstance();
            XPath xpath = xPathFactory.newXPath();
            XPathExpression expr = xpath.compile("*");
            Object result = expr.evaluate(element, XPathConstants.NODESET);
            NodeList nodes = (NodeList) result;
           
            Node node = nodes.item(0);
           
            expr = xpath.compile("*");
            result = expr.evaluate(node, XPathConstants.NODESET);
            nodes = (NodeList) result;
           
            node = nodes.item(0);
            attributes = node.getAttributes();
            if (attributes.getNamedItem("base") != null){
                type = attributes.getNamedItem("base").getNodeValue();
            }
        }
       
        type = convertType(type);
       
        return type;
    }
    
    /**
     * The given node should have the reference to a complexType node. This returns the complexType node.
     * 
     * @param node
     * @return The complexType node
     * @throws XPathExpressionException 
     */
    public Node getStrangeComplex(Node node) throws XPathExpressionException{
        XPathFactory xPathFactory = XPathFactory.newInstance();
        XPath xpath = xPathFactory.newXPath();
        XPathExpression expr = xpath.compile("../*");
       
        Object result = expr.evaluate(node, XPathConstants.NODESET);
        
        NodeList nodes = (NodeList) result;
        int complexPosition = (-1);
        for (int i = 0; i < nodes.getLength(); i++){
            if (nodes.item(i).getNodeName().equals(prefix + ":complexType")){
                complexPosition = i;
            }
        }
        
        return nodes.item(complexPosition);        
    }
    
    /**
     * Finds all the brothers with reference to given complex node
     * 
     * @param complex
     * @return Brothers of given complex node
     * @throws XPathExpressionException 
     */
    public List<Node> getStrangeBrothers(Node complex) throws XPathExpressionException{
        XPathFactory xPathFactory = XPathFactory.newInstance();
        XPath xpath = xPathFactory.newXPath();
        XPathExpression expr = xpath.compile("../*");
       
        Object result = expr.evaluate(complex, XPathConstants.NODESET);
        
        NodeList nodes = (NodeList) result;
        List<Node> resultList = new ArrayList();
        for (int i = 0; i < nodes.getLength(); i++){
            if (!(nodes.item(i).getNodeName().equals(prefix + ":complexType"))){
                resultList.add(nodes.item(i));
            }
        }
        
        return resultList;
    }
   
    /**
     * It converts XML attribute simple type to Java type.
     *
     * @param typeXML XML attribute type
     * @return corresponding Java type - if corresponding type exists;
     *         original type otherwise
     */
    public String convertType(String typeXML){
       
        String typeJava;
       
        HashMap<String, String> typeMap = new HashMap<String, String>();
       
        typeMap.put(prefix + ":string", "java.lang.String");
        typeMap.put(prefix + ":integer", "java.math.BigInteger");
        typeMap.put(prefix + ":int", "int");
        typeMap.put(prefix + ":long", "long");
        typeMap.put(prefix + ":short", "short");
        typeMap.put(prefix + ":decimal", "java.math.BigDecimal");
        typeMap.put(prefix + ":float", "float");
        typeMap.put(prefix + ":double", "double");
        typeMap.put(prefix + ":boolean", "boolean");
        typeMap.put(prefix + ":byte", "byte");
        typeMap.put(prefix + ":QName", "javax.xml.namespace.QName");
        typeMap.put(prefix + ":dateTime", "java.util.Calendar");
        typeMap.put(prefix + ":base64Binary", "byte[]");
        typeMap.put(prefix + ":hexBinary", "byte[]");
        typeMap.put(prefix + ":unsignedInt", "long");
        typeMap.put(prefix + ":unsignedShort", "int");
        typeMap.put(prefix + ":unsignedByte", "short");
        typeMap.put(prefix + ":time", "java.util.Calendar");
        typeMap.put(prefix + ":date", "java.util.Calendar");
        typeMap.put(prefix + ":anyURI", "java.lang.String");
        typeMap.put(prefix + ":anySimpleType", "java.lang.String");
       
        if(typeMap.containsKey(typeXML)){
            typeJava = typeMap.get(typeXML);
        }else{
            typeJava = typeXML;
        }
       
        return typeJava;
    }
   
    /**
     * It returns a node list for given XPath.
     *
     * @param path XPath
     * @return Node list
     */
    public NodeList xPathToNodeList(String path){
        NodeList nodeList = null;
        XPathFactory xPathFactory = XPathFactory.newInstance();
        XPath xpath = xPathFactory.newXPath();
       
        try{
            XPathExpression expr = xpath.compile(path);
            Object result = expr.evaluate(xmlSchema, XPathConstants.NODESET);
            nodeList = (NodeList) result;          
        }catch(XPathExpressionException e){
            System.err.println("XPathExpressionException: " + e.getMessage());
        }
        return nodeList;
    }

    /**
     * Returns the number of occurence of 'name' in 'classNameList'
     *
     * @param name
     * @param classNameList
     * @return The number of occurence
     */
    private int getOccurence(String name, List<String> classNameList) {
        int count = 0;
        for (String className: classNameList){
            if (className.equals(name)){
                count++;
            }
        }
        return count;
    }
}

